package com.huguesjohnson.jail.jailBreak;

/*
jailBreak - Package to launch Java Artificial Intelligence Library components
JailBreakObjectViewerInternalFrame.java - Internal frame which displays detailed information about an object
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.lang.Boolean;
import java.lang.Double;
import java.lang.Object;
import java.lang.String;
import java.awt.Graphics;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import javax.swing.BorderFactory;
import javax.swing.JInternalFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import com.huguesjohnson.jail.jailBreak.JailBreakUIConstants;
import com.huguesjohnson.jail.jailBreak.event.JailBreakEvent;
import com.huguesjohnson.jail.neuralNetwork.BackPropagationNetwork;
import com.huguesjohnson.jail.neuralNetwork.Layer;
import com.huguesjohnson.jail.neuralNetwork.NeuralNetwork;
import com.huguesjohnson.jail.neuralNetwork.Neuron;

/** JailBreakObjectViewerInternalFrame Internal frame which displays detailed information about an object
 * @author Hugues Johnson
 */
public class JailBreakObjectViewerInternalFrame extends JInternalFrame implements ActionListener{

    /** the path to the folder containing images for current theme */
    private String resourcePath;

    /** the path to the folder containing default images */
    private String defaultResourcePath;
    
    /** reference to the settings table */
    private JailBreakTable settingsTable;

    /** reference to the input table */
    private JailBreakTable inputTable;
    
    /** reference to the weights table */
    private JailBreakTable weightsTable;
    
    /** reference to the desired output table */    
    private JailBreakTable desiredOutputTable;
    
    /** reference to the output table */
    private JailBreakTable outputTable;
    
    /** is this the first time this component has been painted? */
    private boolean firstPaint;
    
    /** split pane holding input/output tables */
    private JSplitPane inputOutputSplitPane;
    
    /** split pane holding settings tables */
    private JSplitPane settingsSplitPane;
    
    /** Creates new JailBreakObjectViewerInternalFrame
     * @param resourcePath path to the folder containing images for current theme 
     * @param defaultResourcePath path to the folder containing default images   
     */
    public JailBreakObjectViewerInternalFrame(String resourcePath,String defaultResourcePath){
        super(JailBreakUIConstants.CAPTION_VIEW_FRAME,JailBreakUIConstants.INTERNAL_FRAME_RESIZABLE,JailBreakUIConstants.INTERNAL_FRAME_CLOSABLE,JailBreakUIConstants.INTERNAL_FRAME_MAXIMIZABLE,JailBreakUIConstants.INTERNAL_FRAME_ICONIFIABLE);
        this.firstPaint=true;
        this.resourcePath=resourcePath;
        this.defaultResourcePath=defaultResourcePath;
        this.addComponents();
    }
    
    /** adds components to the frame
     */
    private void addComponents(){
        /* settings panel */
        JPanel settingsPanel=new JPanel(new BorderLayout());
        settingsPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),"Settings"));
        settingsPanel.setLayout(new GridLayout(1,1));
        this.settingsTable=new JailBreakTable(this.resourcePath,this.defaultResourcePath,null,JailBreakUIConstants.SETTINGS_TABLE_COLUMNS);
        JScrollPane scrollPane=new JScrollPane(this.settingsTable);
        settingsPanel.add(scrollPane);
        /* input panel */
        JPanel inputPanel=new JPanel();
        inputPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),"Input"));
        inputPanel.setLayout(new GridLayout(1,2));
        this.inputTable=new JailBreakTable(this.resourcePath,this.defaultResourcePath,null,JailBreakUIConstants.INPUT_TABLE_COLUMNS);
        scrollPane=new JScrollPane(this.inputTable);
        inputPanel.add(scrollPane);
        this.weightsTable=new JailBreakTable(this.resourcePath,this.defaultResourcePath,null,JailBreakUIConstants.WEIGHTS_TABLE_COLUMNS);
        scrollPane=new JScrollPane(this.weightsTable);
        inputPanel.add(scrollPane);
        /* output panel */
        JPanel outputPanel=new JPanel();
        outputPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),"Output"));
        outputPanel.setLayout(new GridLayout(1,2));
        this.desiredOutputTable=new JailBreakTable(this.resourcePath,this.defaultResourcePath,null,JailBreakUIConstants.DESIRED_OUTPUT_TABLE_COLUMNS);
        scrollPane=new JScrollPane(this.desiredOutputTable);
        outputPanel.add(scrollPane);
        this.outputTable=new JailBreakTable(this.resourcePath,this.defaultResourcePath,null,JailBreakUIConstants.OUTPUT_TABLE_COLUMNS);
        scrollPane=new JScrollPane(this.outputTable);
        outputPanel.add(scrollPane);        
        /* create split panes */
        this.inputOutputSplitPane=new JSplitPane(JSplitPane.VERTICAL_SPLIT,inputPanel,outputPanel);
        this.settingsSplitPane=new JSplitPane(JSplitPane.VERTICAL_SPLIT,settingsPanel,this.inputOutputSplitPane);
        /* add split panes */
        this.getContentPane().add(this.settingsSplitPane);
    }
    
    /** paints the component, on first paint this sets the dividers on the split pane to a reasonable proportion
     * param g the Graphics context in which to paint
     */
    public void paint(Graphics g){
        super.paint(g);
        if(this.firstPaint==true){
            this.firstPaint=false;
            /* split panes aren't sized well by default */
            this.inputOutputSplitPane.setDividerLocation(0.5D);
            this.inputOutputSplitPane.setDividerLocation(this.inputOutputSplitPane.getDividerLocation()+this.inputOutputSplitPane.getDividerSize());
            this.settingsSplitPane.setDividerLocation(0.4D);
            this.settingsSplitPane.setDividerLocation(this.settingsSplitPane.getDividerLocation()+this.settingsSplitPane.getDividerSize());
        }
    }
    
    /** updates the frame based on the object passed
     * @param bpNetwork a backpropagation neural network to display
     */
    public void updateFrame(NeuralNetwork network){
        /* populate settings table */
        Object[][] settingsTableRows=JailBreakUIConstants.NETWORK_SETTINGS_ROWS;
        settingsTableRows[JailBreakUIConstants.NETWORK_SETTINGS_ROW_ERROR][1]=new Double(network.getError());
        this.settingsTable.setTableModel(settingsTableRows,JailBreakUIConstants.SETTINGS_TABLE_COLUMNS);
        settingsTableRows=null;
        /* populate input table */
        this.inputTable.setTableModel(null,JailBreakUIConstants.INPUT_TABLE_COLUMNS);
        /* populate weights table */
        this.weightsTable.setTableModel(null,JailBreakUIConstants.WEIGHTS_TABLE_COLUMNS);
        /* populate desired output table */    
        this.desiredOutputTable.setTableModel(null,JailBreakUIConstants.DESIRED_OUTPUT_TABLE_COLUMNS);
        /* populate output table */
        double output[]=network.getOutput();
        int length=output.length;
        Object[][] outputTableRows=new Object[length][2];
        for(int index=0;index<length;index++){
            outputTableRows[index][0]=new String(String.valueOf((int)index));
            outputTableRows[index][1]=new Double(output[index]);
        }
        this.outputTable.setTableModel(outputTableRows,JailBreakUIConstants.OUTPUT_TABLE_COLUMNS);
        outputTableRows=null;
    }    

    /** updates the frame based on the object passed
     * @param bpNetwork a backpropagation neural network to display
     */
    public void updateFrame(Layer layer){
        /* populate settings table */
        Object[][] settingsTableRows=JailBreakUIConstants.LAYER_SETTINGS_ROWS;
        settingsTableRows[JailBreakUIConstants.LAYER_SETTINGS_ROW_ERROR][1]=new Double(layer.getError());
        settingsTableRows[JailBreakUIConstants.LAYER_SETTINGS_ROW_DONE_TRAINING][1]=new Boolean(layer.getDoneTraining());
        this.settingsTable.setTableModel(settingsTableRows,JailBreakUIConstants.SETTINGS_TABLE_COLUMNS);
        settingsTableRows=null;
        /* populate input table */
        this.inputTable.setTableModel(null,JailBreakUIConstants.INPUT_TABLE_COLUMNS);
        /* populate weights table */
        this.weightsTable.setTableModel(null,JailBreakUIConstants.WEIGHTS_TABLE_COLUMNS);
        /* populate desired output table */    
        double desiredOutput[]=layer.getDesiredOutput();
        int length=desiredOutput.length;
        Object[][] desiredOutputTableRows=new Object[length][2];
        for(int index=0;index<length;index++){
            desiredOutputTableRows[index][0]=new String(String.valueOf((int)index));
            desiredOutputTableRows[index][1]=new Double(desiredOutput[index]);
        }
        this.desiredOutputTable.setTableModel(desiredOutputTableRows,JailBreakUIConstants.DESIRED_OUTPUT_TABLE_COLUMNS);
        desiredOutputTableRows=null;
        /* populate output table */
        double output[]=layer.getOutput();
        length=output.length;
        Object[][] outputTableRows=new Object[length][2];
        for(int index=0;index<length;index++){
            outputTableRows[index][0]=new String(String.valueOf((int)index));
            outputTableRows[index][1]=new Double(output[index]);
        }
        this.outputTable.setTableModel(outputTableRows,JailBreakUIConstants.OUTPUT_TABLE_COLUMNS);
        outputTableRows=null;
    }    

    /** updates the frame based on the object passed
     * @param bpNetwork a backpropagation neural network to display
     */
    public void updateFrame(Neuron neuron){
        /* populate settings table */
        Object[][] settingsTableRows=JailBreakUIConstants.NEURON_SETTINGS_ROWS;
        settingsTableRows[JailBreakUIConstants.NEURON_SETTINGS_ROW_DONE_TRAINING][1]=new Boolean(neuron.getDoneTraining());
        settingsTableRows[JailBreakUIConstants.NEURON_SETTINGS_ROW_PUNISH_RATE][1]=new Double(neuron.getPunishRate());
        settingsTableRows[JailBreakUIConstants.NEURON_SETTINGS_ROW_REWARD_RATE][1]=new Double(neuron.getRewardRate());
        settingsTableRows[JailBreakUIConstants.NEURON_SETTINGS_ROW_ERROR_THRESHOLD][1]=new Double(neuron.getErrorThreshold());
        this.settingsTable.setTableModel(settingsTableRows,JailBreakUIConstants.SETTINGS_TABLE_COLUMNS);
        settingsTableRows=null;
        /* populate input table */
        double input[]=neuron.getInput();
        int length=input.length;
        Object[][] inputTableRows=new Object[length][2];
        for(int index=0;index<length;index++){
            inputTableRows[index][0]=new String(String.valueOf((int)index));
            inputTableRows[index][1]=new Double(input[index]);
        }
        this.inputTable.setTableModel(inputTableRows,JailBreakUIConstants.INPUT_TABLE_COLUMNS);
        inputTableRows=null;        
        /* populate weights table */
        double weights[]=neuron.getWeights();
        length=weights.length;
        Object[][] weightsTableRows=new Object[length][2];
        for(int index=0;index<length;index++){
            weightsTableRows[index][0]=new String(String.valueOf((int)index));
            weightsTableRows[index][1]=new Double(weights[index]);
        }
        this.weightsTable.setTableModel(weightsTableRows,JailBreakUIConstants.WEIGHTS_TABLE_COLUMNS);
        weightsTableRows=null;             
        /* populate desired output table */    
        Object[][] desiredOutputTableRows={ {"0",new Double(neuron.getDesiredOutput())} };
        this.desiredOutputTable.setTableModel(desiredOutputTableRows,JailBreakUIConstants.DESIRED_OUTPUT_TABLE_COLUMNS);
        desiredOutputTableRows=null;
        /* populate output table */
        Object[][] outputTableRows={ {"0",new Double(neuron.getDesiredOutput())} };
        this.outputTable.setTableModel(outputTableRows,JailBreakUIConstants.OUTPUT_TABLE_COLUMNS);
        outputTableRows=null;
    }    
    
    /** actionListener receives events from the GUI
     * @param e ActionEvent received by UI
     */
     public void actionPerformed(final ActionEvent e){
     }    
    
     /** sendEvent override when creating a new instance of this object 
      * @param event the event to send to the parent class of this object
      */
    public void sendEvent(JailBreakEvent event){ 
    }

    /** receiveEvent receives JailBreakEvents from objects
      * @param event the event received
      */
    public void receiveEvent(JailBreakEvent event){ 
        /* send all events to the parent class */
        this.sendEvent(event);
    }    
}