package com.huguesjohnson.jail.jdt;

/*
Jdt (Java Decision Tree) - Package to create and display a decision tree
DecisionTree.java - representation of a decision tree, contains training and testing methods
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.io.Serializable;
import java.lang.Cloneable;
import java.lang.CloneNotSupportedException;
import java.lang.Integer;
import java.lang.Math;
import java.lang.Object;
import java.lang.String;
import java.util.Arrays;
import java.util.Vector;
import com.huguesjohnson.jail.jdt.Attribute;
import com.huguesjohnson.jail.jdt.DataInstance;
import com.huguesjohnson.jail.jdt.DecisionTreeNode;
import com.huguesjohnson.LogFile;

/** JdtDecisionTree
 */
public class DecisionTree implements Cloneable,Serializable{
    /* private shared members */
    /** root node of tree */    
    private DecisionTreeNode treeRoot;
    /** log file for decision tree */    
    private LogFile logFile;
    /** whether or not to run in debug mode 
     setting debug to true will create a HUGE log file 
     debug should probably not be used for normal execution */    
    private final boolean debug=false;
     
    /** constructor
     * @param attributes vector of all attributes for data
     * @param trainingData vector of data used to train the tree
     * @param classifications names of all valid classifications
     */
    public DecisionTree(Vector attributes, Vector trainingData, String[] classifications){
        if(debug){
            this.logFile=new LogFile("buildTree.log");
        }
        this.treeRoot=decisionTreeLearn(attributes,trainingData,classifications,this.majorityClassification(trainingData,classifications));
        if(debug){
            this.logFile.close();
        }
    }
    
    /** clone
     * @return clone of object
     */
    public Object clone(){
        try{
            return(super.clone());
        } catch(CloneNotSupportedException cnsx){
            return(null);
        }
    }
    
    /** tries to classify a dataInstance
     * @param dataInstance the dataInstance to try and classify
     * @return classification for dataInstance
     */
    public String classify(DataInstance dataInstance){
        DecisionTreeNode treeNode=this.treeRoot;
        while(!treeNode.isLeaf()&&treeNode!=null){
            double compareValue=dataInstance.getData(treeNode.getAttributeIndex());
            int operator=treeNode.getCompareOperator();
            /* true goes to left child, false goes to right child */
            if(operator==DecisionTreeNode.OP_GREATER_THAN){
                if(compareValue>treeNode.getCompareValue()){
                    treeNode=treeNode.getChildNode(DecisionTreeNode.CHILD_LEFT);
                } else{
                    treeNode=treeNode.getChildNode(DecisionTreeNode.CHILD_RIGHT);
                }
            }else if(operator==DecisionTreeNode.OP_LESS_THAN){
                if(compareValue<treeNode.getCompareValue()){
                    treeNode=treeNode.getChildNode(DecisionTreeNode.CHILD_LEFT);
                } else{
                    treeNode=treeNode.getChildNode(DecisionTreeNode.CHILD_RIGHT);
                }            
            }else{
                if(compareValue==treeNode.getCompareValue()){
                    treeNode=treeNode.getChildNode(DecisionTreeNode.CHILD_LEFT);
                } else{
                    treeNode=treeNode.getChildNode(DecisionTreeNode.CHILD_RIGHT);
                }            
            }
        }
        return(treeNode.getClassification());
    }
    
    /** prunes the tree based on a vector of pruning data
     * @param pruneData data used to prune tree
     */
    public void prune(Vector pruneData){
        /* this implementation of pruning is sketchy at best */
        double bestAccuracy=this.testAccuracy(pruneData);
        double pruneAccuracy;
        Vector openList=new Vector(1,1);
        /* perform a depth first search of the tree, replacing each branch with it's left branch then testing accuracy
         * left branch is used because it contains the majority data, or the "true" data
         * this is probably not the fastest way to prune the tree, but it will work
         */
        if(this.debug){
            this.logFile=new LogFile("pruneTree.log");
            this.logFile.log("bestAccuracy="+bestAccuracy);
        }
        DecisionTreeNode removedNode;
        DecisionTreeNode currentNode;
        DecisionTreeNode leftNode;
        DecisionTreeNode rightNode;
        /* add treeRoot to openList, unless is somehow is a leaf - then this will be a very fast prune */
        if(!this.treeRoot.isLeaf()){
            openList.add(this.treeRoot);
        }
        while(!openList.isEmpty()){
            currentNode=(DecisionTreeNode)openList.remove(0);
            removedNode=(DecisionTreeNode)currentNode.clone();
            leftNode=(DecisionTreeNode)currentNode.getChildNode(DecisionTreeNode.CHILD_LEFT).clone();
            rightNode=(DecisionTreeNode)currentNode.getChildNode(DecisionTreeNode.CHILD_RIGHT).clone();
            if(this.debug){
                this.logFile.log("currentNode="+currentNode.toString());
                this.logFile.log("removedNode="+removedNode.toString());
                this.logFile.log("leftNode="+leftNode.toString());
                this.logFile.log("rightNode="+rightNode.toString());
            }
            /* try right node first */
            if(rightNode!=null){
                //currentNode=(JdtDecisionTreeNode)rightNode.clone();
                if(!currentNode.isRoot()){
                    currentNode.getParentNode().setChildNode(DecisionTreeNode.CHILD_RIGHT,rightNode);
                }
                pruneAccuracy=this.testAccuracy(pruneData);
                if(this.debug){
                    this.logFile.log("pruneAccuracy="+pruneAccuracy);
                }
                if(pruneAccuracy>=bestAccuracy){
                    bestAccuracy=pruneAccuracy;
                } else{
                    DecisionTreeNode rightChild=rightNode.getChildNode(DecisionTreeNode.CHILD_RIGHT);
                    if(rightChild!=null){
                        if(!rightChild.isLeaf()){
                            openList.add(rightChild);
                        }
                    }
                    DecisionTreeNode leftChild=rightNode.getChildNode(DecisionTreeNode.CHILD_LEFT);
                    if(leftChild!=null){
                        if(!leftChild.isLeaf()){
                            openList.add(leftChild);
                        }
                    }
                }
            }
            /* now try left node */
            if(leftNode!=null){
                //currentNode=(JdtDecisionTreeNode)leftNode.clone();
                if(!currentNode.isRoot()){
                    currentNode.getParentNode().setChildNode(DecisionTreeNode.CHILD_LEFT,leftNode);
                }
                pruneAccuracy=this.testAccuracy(pruneData);
                if(this.debug){
                    this.logFile.log("pruneAccuracy="+pruneAccuracy);
                }
                if(pruneAccuracy>=bestAccuracy){
                    bestAccuracy=pruneAccuracy;
                } else{
                    DecisionTreeNode rightChild=leftNode.getChildNode(DecisionTreeNode.CHILD_RIGHT);
                    if(rightChild!=null){
                        if(!rightChild.isLeaf()){
                            openList.add(rightChild);
                        }
                    }
                    DecisionTreeNode leftChild=leftNode.getChildNode(DecisionTreeNode.CHILD_LEFT);
                    if(leftChild!=null){
                        if(!leftChild.isLeaf()){
                            openList.add(leftChild);
                        }
                    }
                }
            }
        }
        if(this.debug){
            this.logFile.close();
        }
    }
    
    /** tests the accuracy of the tree
     * @param testData data to use in test
     * @return accuracy of tree
     */
    public double testAccuracy(Vector testData){
        int correctClassCount=0;
        int dataCount=testData.size();
        for(int index=0;index<dataCount;index++){
            DataInstance dataInstance=(DataInstance)testData.get(index);
            String returnClass=this.classify(dataInstance);
            String actualClass=dataInstance.getClassification();
            if(actualClass.equals(returnClass)){
                correctClassCount++;
            }
        }
        return(((double)correctClassCount/(double)dataCount)*100.0D);
    }
    
    /** constructs a decision tree
     * based on the following Decision-Tree-Learn algorithm
     * from: "Artificial Intelligence a Modern Approach" (pg.547) Russell & Norvig
     * DecisionTree-Learn(example,attributes,default)
     * if examples is empty then return default
     * elseif all example have same classification the return classificiation
     * elseif attributes is empty then return majority classification
     * else
     *    best=chooseAttribute(attributes,examples)
     *    tree=new decision-tree with root node=best
     *    for each value vi of best
     *        examplesi=elements of examples with best=vi
     *        subtree=Decision-Tree-Learn(examplesi,attributes-best,default)
     *        add subtree to tree
     *    next
     * return tree
     * @return tree built from learning algorithm
     * @param defaultClass value that is given to data that can not be classified
     * @param attributes vector of all attributes for data
     * @param trainingData vector of data used to train the tree
     * @param classifications names of all valid classifications
     */    
    private DecisionTreeNode decisionTreeLearn(Vector attributes,Vector trainingData,String classifications[],String defaultClass){
        if(debug){this.logFile.log("Enter decisionTreeLearn");}
        DecisionTreeNode treeNode=new DecisionTreeNode();        
        if(trainingData.isEmpty()){
            treeNode.setClassification(defaultClass);
        } else if(this.sameClass(trainingData)){
            treeNode.setClassification(((DataInstance)trainingData.get(0)).getClassification());
        } else if(emptyAttributes(attributes)){
            treeNode.setClassification(this.majorityClassification(trainingData,classifications));
        }else{
            /* here's the tricky part, one slip up and infinite recursion may occur */
            treeNode=this.findBestAttribute(attributes,trainingData,classifications);
            double compareValue=treeNode.getCompareValue();
            int attributeIndex=treeNode.getAttributeIndex();
            int setSize=trainingData.size();
            if(debug){
                this.logFile.log("setSize="+setSize);
            }
            Vector trueData=new Vector(setSize);
            Vector falseData=new Vector(setSize);
            /* can't have negative set size, set size of 0 will cause a divide by zero error */
            if(setSize>0){
                /* split training data into "true" group and "false" group */
                for(int index=0;index<setSize;index++){
                    DataInstance dataInstance=(DataInstance)trainingData.get(index);
                    double value=dataInstance.getData(attributeIndex);
                    if(value>compareValue){
                        trueData.add(dataInstance);
                    } else{
                        falseData.add(dataInstance);
                    }
                }
            }
            /* done copying data, trim vectors */
            trueData.trimToSize();
            falseData.trimToSize();
            if(debug){
                this.logFile.log("trueData.size()="+trueData.size());
                this.logFile.log("falseData.size()="+falseData.size());
            }
            /* create attributesi vector */
            int attributeCount=attributes.size();
            Vector attributesi=new Vector(attributeCount);
            for(int index=0;index<attributeCount;index++){
                /* mark currently used attribute */
                if(index==attributeIndex){
                    attributesi.add(new Attribute("USED",0.0D));
                }else{
                    attributesi.add(attributes.get(index));
                }
            }
            /* here comes the "fun" part, hope there's a lot of free memory */
            treeNode.setChildNode(DecisionTreeNode.CHILD_LEFT,this.decisionTreeLearn(attributesi,trueData,classifications,this.majorityClassification(trueData,classifications)));
            treeNode.setChildNode(DecisionTreeNode.CHILD_RIGHT,this.decisionTreeLearn(attributesi,falseData,classifications,this.majorityClassification(falseData,classifications)));
        }
        if(debug){
            this.logFile.log("Exit decisionTreeLearn with following values:");
            this.logFile.log("treeNode.getClassification()="+treeNode.getClassification());
            this.logFile.log("treeNode.getAttributeIndex()="+treeNode.getAttributeIndex());
            this.logFile.log("treeNode.getCompareValue="+treeNode.getCompareValue());
        }
        return(treeNode);
    }
    
    /** finds attribute with highest gain
     * @param attributes vector of all attributes for data
     * @param trainingData vector of data used to train the tree
     * @param classifications names of all valid classifications
     * @return treenode storing the attributeIndex & compare value for the best attribute test
     */        
    private DecisionTreeNode findBestAttribute(Vector attributes,Vector trainingData,String classifications[]){
        DecisionTreeNode bestNode=new DecisionTreeNode();
        int attributeIndex;
        int attributeCount=attributes.size();
        int bestAttributeIndex=-1;
        double bestGain=0.0D;
        double bestCompareValue=0.0D;
        int progressThreshold;
        if(attributeCount<10){
            progressThreshold=1;
        }else if(attributeCount<100){
            progressThreshold=5;
        }else{
            progressThreshold=attributeCount/10;
        }
        //System.out.print("Finding midpoints in data..");
        for(attributeIndex=0;attributeIndex<attributeCount;attributeIndex++){
            /* display progress */
            if(attributeIndex%progressThreshold==0){
                //System.out.print(this.percentComplete(attributeIndex,attributeCount));
            }
            /* skip over used attributes or else face infinite recursion */
            if(((Attribute)attributes.get(attributeIndex)).getName().equals("USED")==false){
                double midPoints[]=findMidPoints(trainingData,attributeIndex);
                for(int j=0;j<midPoints.length;j++){
                    if(debug){
                        this.logFile.log("attributeIndex="+attributeIndex+"  midPoints["+j+"]="+midPoints[j]);
                    }
                    double g=this.gain(attributeIndex,trainingData,classifications,midPoints[j]);
                    if(g>bestGain){
                        bestGain=g;
                        bestCompareValue=midPoints[j];
                        bestAttributeIndex=attributeIndex;
                    }
                }
            }
        }
        //System.out.print("100%\n");
        /* set values for bestNode */
        bestNode.setCompareValue(bestCompareValue);
        bestNode.setAttributeIndex(bestAttributeIndex);
        bestNode.setCompareOperator(DecisionTreeNode.OP_GREATER_THAN);
        if(debug){
            this.logFile.log("        exit find best node with:");
            this.logFile.log("        bestGain="+bestGain);
            this.logFile.log("        bestAttributeIndex="+bestAttributeIndex);
            this.logFile.log("        bestCompareValue="+bestCompareValue);
        }
        return(bestNode);
    }
    
    /** determines if a set of dataInstances are all the same classification
     * @return true is all attributes are the same class
     * @param dataInstances vector of data instances to compare
     */
    private boolean sameClass(Vector dataInstances){
        boolean same=true;
        int index=0;
        int count=dataInstances.size();
        if(count<=0){
            same=false;
        } else{
            /* get classification of first item and compare to rest of items */
            DataInstance data=(DataInstance)dataInstances.get(index);
            String classification=data.getClassification();
            index++;
            while(same&&index<count){
                data=(DataInstance)dataInstances.get(index);
                if(classification.equals(data.getClassification())==false){
                    same=false;
                } else{
                    index++;
                }
            }
        }
        return(same);
    }
    
    /** computes the entropy for a set
     * @param classifications array of valid classifications for data
     * @param trainingData vector of data used to train tree
     * @return entropy, double is used to prevent ties
     */
    private double entropy(Vector trainingData,String classifications[]){
        double e=0.0D;
        int setSize=trainingData.size();
        int classificationCount=classifications.length;
        int classOccurences[]=new int[classificationCount];
        int index;
        /* set all classification counts to 0 */
        for(index=0;index<classificationCount;index++){
            classOccurences[index]=0;
        }
        /* count how many times each classification occurs in set */
        for(index=0;index<setSize;index++){
            DataInstance dataInstance=(DataInstance)trainingData.get(index);
            int classificationIndex=this.getClassificationIndex(classifications,dataInstance.getClassification());
            if(classificationIndex>=0&&classificationIndex<classificationCount){
                classOccurences[classificationIndex]++;
            }
        }
        /* compute entropy */
        for(index=0;index<classificationCount;index++){
            /* can't take the log of a number <=0 */
            if(classOccurences[index]>0){
                double ratio=(Math.abs((double)classOccurences[index])/Math.abs((double)setSize));
                e-=ratio*logBase2(ratio);
            }            
        }
        return(e);
    }
    
    /** computes the gain for an attribute
     * @param attributeIndex index of attribute to test
     * @param classifications array of valid classifications for data
     * @param trainingData vector of data used to train tree
     * @param compareValue value to test against
     * @return gain, double is used to prevent ties
     */    
    private double gain(int attributeIndex,Vector trainingData,String classifications[],double compareValue){
        double g=0.0D;
        double e=this.entropy(trainingData,classifications);
        int setSize=trainingData.size();
        if(this.debug){
            this.logFile.log("    Enter gain with: attributeIndex="+attributeIndex+" compareValue="+compareValue);
            this.logFile.log("    setSize="+setSize);
            this.logFile.log("    e="+e);
        }
        /* can't have negative set size, set size of 0 will cause a divide by zero error */
        if(setSize>0){
            Vector trueData=new Vector(setSize);
            Vector falseData=new Vector(setSize);
            /* split training data into "true" group and "false" group */
            for(int index=0;index<setSize;index++){
                DataInstance dataInstance=(DataInstance)trainingData.get(index);
                double value=dataInstance.getData(attributeIndex);
                if(value>compareValue){
                    trueData.add(dataInstance);
                } else{
                    falseData.add(dataInstance);
                }
            }
            /* done copying data, trim vectors */
            trueData.trimToSize();
            falseData.trimToSize();
            int trueDataSize=trueData.size();
            int falseDataSize=falseData.size();
            if(this.debug){
                this.logFile.log("    trueDataSize="+trueDataSize);
                this.logFile.log("    falseDataSize="+falseDataSize);
            }
            /* if either vector is empty then gain = entropy of set and no more computations are needed */
            if((trueDataSize==0)||(falseDataSize==0)){
                g=e;
            } else{
                double trueEntropy=this.entropy(trueData,classifications);
                double falseEntropy=this.entropy(falseData,classifications);
                double trueRatio=Math.abs((double)trueDataSize)/Math.abs((double)setSize);
                double falseRatio=Math.abs((double)falseDataSize)/Math.abs((double)setSize);
                if(this.debug){
                    this.logFile.log("    trueEntropy="+trueEntropy);
                    this.logFile.log("    falseEntropy="+falseEntropy);
                    this.logFile.log("    trueRatio="+trueRatio);
                    this.logFile.log("    falseRatio="+falseRatio);
                }
                /* finally, we can compute the gain */
                g=e-(trueRatio*trueEntropy)-(falseRatio*falseEntropy);
            }
        } 
        if(this.debug){
            this.logFile.log("    Exit gain with e="+e+" g="+g);
        }
        return(g);
    }
    
    /** retrieves index of a classification from array of classifications
     * @return index of search string
     * @param classifications array of valid classifications
     * @param search string to search for in array
     */
    private int getClassificationIndex(String classifications[],String search){
        int index=0;
        int foundIndex=-1;
        while((index<classifications.length)&&(foundIndex<0)){
            if(classifications[index].equals(search)){
                foundIndex=index;
            } else{
                index++;
            }
        }
        return(foundIndex);
    }
    
    /** calculates the midpoints of all data stored in a set of data
     * @param trainingData set of data
     * @param attributeIndex item in dataInstance to use to calculate midpoints
     * @return array of midpoints
     */
    private double[] findMidPoints(Vector trainingData,int attributeIndex){
        int index;
        if(trainingData.size()<1){
            return(new double[0]);
        } else{
            int size=trainingData.size();
            double data[]=new double[size];
            /* copy data from dataInstance to data[] */
            for(index=0;index<size;index++){
                data[index]=((DataInstance)trainingData.get(index)).getData(attributeIndex);
            }
            Arrays.sort(data);
            double midPoints[]=new double[size-1];
            for(index=0;index<size-1;index++){
                midPoints[index]=(data[index]+data[index+1])/2.0D;
            }
            return(midPoints);
        }
    }
    
    /** tests if every attribute is used
     * @param attributes vector of attributes to test
     * @return true if all attributes are used
     */
    private boolean emptyAttributes(Vector attributes){
        int index=0;
        int size=attributes.size();
        boolean empty=true;
        while(empty&&(index<size)){
            Attribute attribute=(Attribute)attributes.get(index);
            if(attribute.getName().equals("USED")==false){
                empty=false;
            }else{
                index++;
            }
        }
        return(empty);
    }
    
    /** majorityClassification, finds the majority class for a set of data
     * @param data data to search
     * @param classifications array of valid classifications
     * @return name of majority class
     */
    private String majorityClassification(Vector data,String classifications[]){
        int majorityIndex=0;
        int setSize=data.size();
        int classificationCount=classifications.length;
        int classOccurences[]=new int[classificationCount];
        int index;
        /* set all classification counts to 0 */
        for(index=0;index<classificationCount;index++){
            classOccurences[index]=0;
        }
        /* count how many times each classification occurs in set */
        for(index=0;index<setSize;index++){
            DataInstance dataInstance=(DataInstance)data.get(index);
            int classificationIndex=this.getClassificationIndex(classifications,dataInstance.getClassification());
            if(classificationIndex>=0&&classificationIndex<classificationCount){
                classOccurences[classificationIndex]++;
            }
        }
        /* majorityIndex starts at 0, begin this loop at 1 */
        for(index=1;index<classificationCount;index++){
            if(classOccurences[index]>classOccurences[majorityIndex]){
                majorityIndex=index;
            }
        }
        return(classifications[majorityIndex]);
    }
    
    /** returns a string representing a breadth first search of the tree
     * @return breadth first search of tree
     */
    public String breadthFirstSearch(){
        String bfs=new String();
        Vector openList=new Vector(1,1);
        openList.add(this.treeRoot);
        while(!openList.isEmpty()){
            DecisionTreeNode treeNode=(DecisionTreeNode)openList.remove(0);
            bfs=bfs.concat(treeNode.toString()+" ");
            DecisionTreeNode childNode=treeNode.getChildNode(DecisionTreeNode.CHILD_LEFT);
            if(childNode!=null){
                openList.add(childNode);
            }
            childNode=treeNode.getChildNode(DecisionTreeNode.CHILD_RIGHT);
            if(childNode!=null){
                openList.add(childNode);
            }
        }        
        return(bfs);
    }

    /** returns a vector representing a breadth first search of the tree
     * @return breadth first search of tree
     */
    public Vector breadthFirstSearchVector(){
        Vector bfs=new Vector();
        Vector openList=new Vector(1,1);
        openList.add(this.treeRoot);
        while(!openList.isEmpty()){
            DecisionTreeNode treeNode=(DecisionTreeNode)openList.remove(0);
            bfs.add(treeNode);
            DecisionTreeNode childNode=treeNode.getChildNode(DecisionTreeNode.CHILD_LEFT);
            if(childNode!=null){
                openList.add(childNode);
            }
            childNode=treeNode.getChildNode(DecisionTreeNode.CHILD_RIGHT);
            if(childNode!=null){
                openList.add(childNode);
            }
        }        
        return(bfs);
    }
    
    /** returns a string representation of the tree
     * @return representation of the tree
     */
    public String toString(){
        Vector expandedList=new Vector();
        Vector openList=new Vector();
        Vector depthList=new Vector();
        Vector stringVector=new Vector();
        DecisionTreeNode treeNode;
        String treeToString=new String("");
        openList.add(this.treeRoot); 
        String currentLine=new String(this.treeRoot.toString());
        int maxDepth=-1;
        int currentDepth=0;
        depthList.add(new Integer(currentDepth));
        while(!openList.isEmpty()){
            treeNode=(DecisionTreeNode)openList.remove(0);
            currentDepth=((Integer)(depthList.remove(0))).intValue();
            if(currentDepth>maxDepth){
                if(currentDepth<10){
                    currentLine="depth 0"+currentDepth+": "+currentLine;
                }
                else{
                    currentLine="depth "+currentDepth+": "+currentLine;
                }
                stringVector.add(currentLine);
                currentLine=new String("");
                maxDepth=currentDepth;
                //treeToString=treeToString.concat("\n");
            } 
            DecisionTreeNode left=treeNode.getChildNode(DecisionTreeNode.CHILD_LEFT);
            DecisionTreeNode right=treeNode.getChildNode(DecisionTreeNode.CHILD_RIGHT);
            if(left==null){
                //currentLine=currentLine.concat(" ");
            } else{
                currentLine=currentLine.concat(left.toString()+"     ");
            }
            if(right==null){
                //currentLine=currentLine.concat(blank);
            } else{
                currentLine=currentLine.concat(right.toString()+"     ");
            }
            if(treeNode.getChildNode(DecisionTreeNode.CHILD_LEFT)!=null){
                openList.add(treeNode.getChildNode(DecisionTreeNode.CHILD_LEFT));
                depthList.add(new Integer(currentDepth+1));
            }
            if(treeNode.getChildNode(DecisionTreeNode.CHILD_RIGHT)!=null){
                openList.add(treeNode.getChildNode(DecisionTreeNode.CHILD_RIGHT));
                depthList.add(new Integer(currentDepth+1));
            }
            openList.trimToSize();
        }
        /* copy contents of stringVector into treeToString */
        stringVector.trimToSize();
        //int longestLine=(int)Math.pow(2,maxDepth)*nodeLength;
        while(!stringVector.isEmpty()){
            currentLine=(String)stringVector.remove(0);
            //currentLine=this.fitTo(currentLine,longestLine);
            treeToString=treeToString+"\n"+currentLine;
        }
        //System.out.println("\n\nmaxDepth="+maxDepth+"\n\n");
        //System.out.println("\n\nlongest line="+((int)Math.pow(2,maxDepth)*nodeLength+"\n\n"));
        return(treeToString);
    }
    
    /** builds a string reprenting the current progress of a loop
     * @param current current index
     * @param max upper bound of for loop
     * @return string representing progress
     */
    private String percentComplete(int current,int max){
        int percentComplete;
        if(max==0){
            percentComplete=0;
        } else{
            percentComplete=(int)((float)current/(float)max*100.0F);
        }
        return(new String(percentComplete+"% "));
    }
    
    /** computes log base 2 of a number
     * why is this not in the java.lang.Math class?
     * at the very least you'd expect a logBaseN() method..
     * @param n number to take log base 2 of
     * @return log base 2 of n
     */
    private final double logBase2(double n){
        if(n<0){
            return(0.0D);
        } else{
            return(Math.log(n)/Math.log(2.0D));
        }
    }
}