package com.huguesjohnson.jail.neuralNetwork;

/*
neuralNetwork - Package of neural network objects
BackPropagationNetwork.java - Backpropagation neural network
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import com.huguesjohnson.jail.neuralNetwork.BackPropagationLayer;
import com.huguesjohnson.jail.neuralNetwork.LayerException;
import com.huguesjohnson.jail.neuralNetwork.NeuralNetwork;
import com.huguesjohnson.jail.neuralNetwork.NeuralNetworkException;
import com.huguesjohnson.jail.neuralNetwork.Perceptron;

/** BackPropagationNetwork
 */
public class BackPropagationNetwork extends NeuralNetwork{
    /** the input layer for the network */
    private BackPropagationLayer inputLayer;
    
    /** the hidden layer for the network */
    private BackPropagationLayer hiddenLayer;
    
    /** the output layer for the network */
    private BackPropagationLayer outputLayer;
    
    /* constructor(s) */

    /** constructor
     * @param inputLayerSize the number of neurons in the input layer
     * @param hiddenLayerSize the number of neurons in the hidden layer
     * @param outputLayerSize the number of neurons in the output layer
     * @param inputSize the number of inputs to the layer
     * @param punishRate the punish rate for every neuron in the layer
     * @param rewardRate the reward rate for every neuron in the layer
     * @param errorThreshold the error threshold for every neuron in the layer
     * @param minWeight minimum weight on a neuron link
     * @param maxWeight maximum weight on a neuron link
     */    
    public BackPropagationNetwork(int inputLayerSize,int hiddenLayerSize,int outputLayerSize,int inputSize,double punishRate,double rewardRate,double errorThreshold,double minWeight,double maxWeight){
        super();
        /* create input layer */
        this.inputLayer=new BackPropagationLayer(inputLayerSize,inputSize,punishRate,rewardRate,errorThreshold,1.0D,1.0D){
            public void runComplete(){
                runHiddenLayer();
            }
        };
        /* links from input to inputLayer are not trained */
        this.inputLayer.setDoneTraining(true);
        /* create hidden layer */
        this.hiddenLayer=new BackPropagationLayer(hiddenLayerSize,inputLayerSize,punishRate,rewardRate,errorThreshold,minWeight,maxWeight){
            public void runComplete(){
                runOutputLayer();
            }
        };
        /* create output layer */
        this.outputLayer=new BackPropagationLayer(outputLayerSize,hiddenLayerSize,punishRate,rewardRate,errorThreshold,minWeight,maxWeight){
            public void runComplete(){
                networkRunComplete();
            }
        };
    }
      
    /* overrides of NeuralNetwork methods */
    
    /** returns the output of the network
     * @return output of the network
     */
    public double[] getOutput(){
        return(this.outputLayer.getOutput());
    }

    /** returns the error of the network
     * @return error of the network
     */
    public double getError(){
        return(this.outputLayer.getError());
    }
    
    /** sets the desired output of the network
     * @param desiredOutput the desired output of the network
     */
    public void setDesiredOutput(double[] desiredOutput){
        try{
            this.outputLayer.setDesiredOutput(desiredOutput);
        } catch(LayerException lx){
            this.log(lx.toString());
            lx.printStackTrace();
        }
    }
    
    /** sets the input of the network
     * @param input the input of the network
     */
    public void setInput(double[] input){
        this.inputLayer.setInput(input);
    }

    /** returns the input layer 
     * @return input layer
     */
    public final BackPropagationLayer getInputLayer(){
        return(this.inputLayer);
    }
    
    /** returns the hidden layer 
     * @return hidden layer
     */
    public final BackPropagationLayer getHiddenLayer(){
        return(this.hiddenLayer);
    }

    /** returns the output layer 
     * @return output layer
     */
    public final BackPropagationLayer getOutputLayer(){
        return(this.outputLayer);
    }

    /** sets the input layer, automatically adjusts links between input layer and hidden layer
     * @param inputLayer new input layer
     */
    public final void setInputLayer(BackPropagationLayer inputLayer){
        this.inputLayer=inputLayer;
    }

    /** sets the hidden layer, automatically adjusts links between input layer and hidden layer and hidden layer and output layer 
     * @param hiddenLayer new hidden layer
     */
    public final void setHiddenLayer(BackPropagationLayer hiddenLayer){
        this.hiddenLayer=hiddenLayer;
    }    
    
    /** sets the output layer, automatically adjusts links between hidden layer and output layer 
     * @param outputLayer new output layer
     */
    public final void setOutputLayer(BackPropagationLayer outputLayer){
        this.outputLayer=outputLayer;
    }    
    
    /** train the network
     * @param iterations number of training iterations to run
     */
    public void train(int iterations){
        this.log("");
        this.log("enter train()");
        this.log("iterations="+iterations);
        /* reset training status of all layers */
        this.hiddenLayer.setDoneTraining(false);
        this.outputLayer.setDoneTraining(false);
        /* start training the network */
        for(int index=0;index<iterations;index++){
            double startError=this.getError();
            /* try to train the network */
            if(this.outputLayer.getDoneTraining()==false){
                this.trainOutputLayer();
            } else if(this.hiddenLayer.getDoneTraining()==false){
                this.trainHiddenLayer();
            } else{
                this.log("all layers done training, decreasing training rate..");
                this.inputLayer.adjustTrainingRate(0.9);
                this.hiddenLayer.adjustTrainingRate(0.9);
                this.outputLayer.adjustTrainingRate(0.9);
                this.trainOutputLayer();
            }
            this.run();
            double newError=this.getError();
            if(newError==startError){
                this.log("NO IMPROVEMENT! newError ("+newError+") == startError ("+startError+")");
            } else if(newError<startError){
                this.log("PASS! newError ("+newError+") < startError ("+startError+")");
            } else{ /* newError>startError */
                this.log("FAIL! newError ("+newError+") > startError ("+startError+")");
            }
        }
        this.log("exit train()");
    }
    
    /** run the network
     */
    public void run(){
        this.runInputLayer();
    }
    
    /* methods unique to BackPropagationNetwork */

    /** run the input layer
     */
    private final void runInputLayer(){
        this.inputLayer.run();
    }
    
    /** run the hidden layer
     */
    private final void runHiddenLayer(){
        this.hiddenLayer.setInput(this.inputLayer.getOutput());
        this.hiddenLayer.run();
    }
    
    /** run the output layer
     */
    private final void runOutputLayer(){
        this.outputLayer.setInput(this.hiddenLayer.getOutput());
        this.outputLayer.run();
    }
    
    /** train the hidden layer
     */
    private final void trainHiddenLayer(){
        this.log("enter trainHiddenLayer()");
        /* need to determine desired output for hidden layer */
        double[] hiddenLayerOutput=this.hiddenLayer.getOutput();
        int length=hiddenLayerOutput.length;
        /* manually copy output to hiddenLayerDesiredOutput */
        double[] hiddenLayerDesiredOutput=new double[length];
        for(int index=0;index<length;index++){
            hiddenLayerDesiredOutput[index]=hiddenLayerOutput[index];
        }
        /* find desired output which will most improve output of layer */
        double startError=this.outputLayer.getError();
        double bestError=startError;
        for(int index=0;index<length;index++){
            double originalValue=hiddenLayerDesiredOutput[index];
            if(originalValue>0.0D){
                hiddenLayerDesiredOutput[index]=-1.0D;
            } else{
                hiddenLayerDesiredOutput[index]=1.0D;
            }
            /* try running the new input */
            this.outputLayer.setInput(hiddenLayerDesiredOutput);
            this.outputLayer.run();
            /* if no improvement was made, undo change */
            if(this.outputLayer.getError()>=bestError){
                hiddenLayerDesiredOutput[index]=originalValue;
            } else{
                bestError=this.outputLayer.getError();
            }
        }
        if(bestError<startError){
            try{
                this.log("bestError<startError - set desired output for hidden layer and train");
                hiddenLayer.setDesiredOutput(hiddenLayerDesiredOutput);
                hiddenLayer.train();
            } catch(LayerException lx){
                lx.printStackTrace();
                this.log("EXCEPTION: "+lx.toString());
            }
        } else{
            this.log("bestError>=startError - improvement for hiddenLayer stopped, setting hiddenLayer.doneTraining=true..");
            this.hiddenLayer.setDoneTraining(true);
        }
        this.log("exit trainHiddenLayer()");
    }

    /** train the output layer
     */
    private final void trainOutputLayer(){
        this.log("enter trainOutputLayer()");
        /* output layer knows it's desired output, train as usual */
        this.outputLayer.train();
        this.log("exit trainOutputLayer()");
    }
    
    /** method invoked when the network run is complete
     */
    public void networkRunComplete(){
        /* override in implementing class */
    }
    
    /** returns the desired output of the network
     * @return desired output of the network
     */
    public double[] getDesiredOutput(){
        return(this.outputLayer.getDesiredOutput());
    }        
}