package com.huguesjohnson.jail.neuralNetwork;

/*
neuralNetwork - Package of neural network objects
KohonenLayer.java - Kohonen self-organizing map layer
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.lang.Double;
import java.lang.String;
import com.huguesjohnson.jail.neuralNetwork.KohonenNeuron;
import com.huguesjohnson.jail.neuralNetwork.Layer;
import com.huguesjohnson.jail.neuralNetwork.LayerException;

/** KohonenLayer
 */
public class KohonenLayer extends Layer{

    /* contructor(s) */
    
    /** default constructor
     */
    public KohonenLayer(){
        super();
    }
    
    /** constructor
     * @param layerSize the number of neurons in the layer
     * @param inputSize the number of inputs to the layer
     * @param punishRate the punish rate for every neuron in the layer
     * @param rewardRate the reward rate for every neuron in the layer
     * @param errorThreshold the error threshold for every neuron in the layer
     */    
    public KohonenLayer(int layerSize,int inputSize,double punishRate,double rewardRate,double errorThreshold){
        super();
        for(int index=0;index<layerSize;index++){
            KohonenNeuron neuron=new KohonenNeuron(punishRate,rewardRate,errorThreshold);
            neuron.randomizeWeights(inputSize,-1.0D,1.0D);
            this.insertNeuron(neuron);
        }
    }
    
    /* overrides of Layer methods */

    /** method invoked when run is complete
     */
    public void runComplete(){
        /* override this in implementing class */
    }


    /** move actual output closer to desired output 
     */
    public void train(){
        double error=this.getError();
        if(error>0.0D){
            /* find which neuron in this layer is supposed to fire */
            double[] desiredOutput=this.getDesiredOutput();
            int length=desiredOutput.length;
            int fireIndex=-1;
            int index=0;
            while((index<length)&&(fireIndex<0)){
                if(desiredOutput[index]==1.0D){
                    fireIndex=index;
                }
                index++;
            }
            /* train the neuron that's supposed to fire */
            try{
                Neuron neuron=this.getNeuron(fireIndex);
                neuron.adjustWeights();
                this.setNeuron(fireIndex,neuron);
            } catch(LayerException lx){
                lx.printStackTrace();
            }
            this.run();
            /* train neurons have a higher activation than the neuron that's supposed to fire */
            error=this.getError();
            if(error>0.0D){
                index=0;
                try{
                    double fireActivation=this.getNeuron(fireIndex).activation();
                    while((index<length)&&(error>0.0D)){
                        if(index!=fireIndex){
                            Neuron neuron=this.getNeuron(index);
                            double activation=neuron.activation();
                            if(activation>=fireActivation){
                                neuron.adjustWeights();
                                this.setNeuron(index,neuron);
                            }
                            this.run();
                            error=this.getError();
                        }
                        index++;
                    }
                } catch(LayerException lx){

                }
            }
        }
    }    
    
    /** compute the output for the layer
     */
    public void computeOutput(){
        double highActivation=0.0D;
        int highActivationIndex=0;
        int size=this.getSize();
        this.initOutput();
        try{
            for(int index=0;index<size;index++){
                double activation=this.getNeuron(index).activation();
                if(activation>highActivation){
                    highActivation=activation;
                    highActivationIndex=index;
                }
            }
            for(int index=0;index<size;index++){
                this.setOutput(index,-1.0D);
            }
            this.setOutput(highActivationIndex,1.0D);
        } catch(LayerException lx){
            lx.printStackTrace();
        }
    }
    
    /* methods specific to KohonenLayer */
    
}