package com.huguesjohnson.jail.neuralNetwork;

/*
neuralNetwork - Package of neural network objects
Layer.java - Abstract base class for all layers
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.io.Serializable;
import java.lang.ArrayIndexOutOfBoundsException;
import java.lang.Math;
import java.lang.Object;
import java.lang.Runnable;
import java.lang.String;
import java.lang.StringBuffer;
import java.lang.Thread;
import java.util.Vector;
import com.huguesjohnson.jail.neuralNetwork.LayerException;
import com.huguesjohnson.jail.neuralNetwork.Neuron;

/** class Layer
 */
public abstract class Layer extends Object implements Runnable,Serializable{
    /** vector of neurons stored in the layer */
    private Vector neurons;
    
    /** the output of the layer */
    private double[] output;
    
    /** error of the layer */
    private double error;
    
    /** thread for this class */
    private Thread myThread;
    
    /* contructor(s) */
    
    /** default constructor
     */
    public Layer(){
        super();
        this.neurons=new Vector();
        this.output=new double[0];
        this.error=-1.0D;
    }
    
    /* methods common to all Layer objects that should not be overridden */
    
    /** sets the input to the layer
     * each neuron in the layer receives the same input
     * @param input the input to the layer
     */
    public final void setInput(double[] input){
        int size=this.neurons.size();
        for(int index=0;index<size;index++){
            ((Neuron)this.neurons.elementAt(index)).setInput(input);
        }
    }
    
    /** sets the desired output of the layer, needed for training
     * @param desiredOutput the desired output of the layer
     * @throws LayerException if the size of desiredOutput is not equal to the number of neurons in the layer
     */
    public final void setDesiredOutput(double[] desiredOutput) throws LayerException{
        int size=this.neurons.size();
        if(size!=desiredOutput.length){
            throw new LayerException("Desired output size ("+desiredOutput.length+") does not match layer size ("+size+")");
        } else{
            for(int index=0;index<size;index++){
                ((Neuron)this.neurons.elementAt(index)).setDesiredOutput(desiredOutput[index]);
            }
        }
    }
    
    /** returns the desired output of the layer
     * @return the desired output of the layer
     */
    public final double[] getDesiredOutput(){
        int size=this.neurons.size();
        if(size<1){
            return(null);
        } else{
            double[] desiredOutput=new double[size];
            for(int index=0;index<size;index++){
                desiredOutput[index]=((Neuron)this.neurons.elementAt(index)).getDesiredOutput();
            }
            return(desiredOutput);
        }
    }    
    
    /** inserts a neuron to the bottom of the layer
     * @param neuron the neuron to add
     */
    public final void insertNeuron(Neuron neuron){
        this.neurons.add(neuron);
    }
    
    /** inserts a neuron at the specified index in the layer
     * @param index the index to insert the neuron at
     * @param neuron the neuron to add
     * @throws LayerException if index<0 or index>=number of neurons in the layer
     */
    public final void insertNeuron(int index,Neuron neuron) throws LayerException{
        if(index<0){
            throw new LayerException("Value passed by index ("+index+") is less than zero");
        } else if(index>=this.neurons.size()){
            throw new LayerException("Value passed by index ("+index+") is greater than or equal to layer size ("+this.neurons.size()+")");
        } else{
            try{
                this.neurons.add(index,neuron);
            } catch(ArrayIndexOutOfBoundsException aiobx){
                throw new LayerException(aiobx.getMessage());
            }
        }
    }
    
    /** removes a neuron at the specified index in the layer
     * @param index the index to remove the neuron from
     * @throws LayerException if index<0 or index>=number of neurons in the layer
     */    
    public final void removeNeuron(int index) throws LayerException{
        if(index<0){
            throw new LayerException("Value passed by index ("+index+") is less than zero");
        } else if(index>=this.neurons.size()){
            throw new LayerException("Value passed by index ("+index+") is greater than or equal to layer size ("+this.neurons.size()+")");
        } else{
            try{
                this.neurons.remove(index);
            } catch(ArrayIndexOutOfBoundsException aiobx){
                throw new LayerException(aiobx.getMessage());
            }
        }
    }

    /** sets a neuron at the specified index in the layer
     * @param index the index to set the neuron at
     * @param neuron the neuron to set
     * @throws LayerException if index<0 or index>=number of neurons in the layer
     */    
    public final void setNeuron(int index,Neuron neuron) throws LayerException{
        if(index<0){
            throw new LayerException("Value passed by index ("+index+") is less than zero");
        } else if(index>=this.neurons.size()){
            throw new LayerException("Value passed by index ("+index+") is greater than or equal to layer size ("+this.neurons.size()+")");
        } else{
            try{
                this.neurons.set(index,neuron);
            } catch(ArrayIndexOutOfBoundsException aiobx){
                throw new LayerException(aiobx.getMessage());
            }
        }
    }    
    
    /** returns the neuron at the specified index in the layer
     * @param index the index of the neuron to return
     * @return the neuron at index
     * @throws LayerException if index<0 or index>=number of neurons in the layer
     */
    public final Neuron getNeuron(int index) throws LayerException{
        if(index<0){
            throw new LayerException("Value passed by index ("+index+") is less than zero");
        } else if(index>=this.neurons.size()){
            throw new LayerException("Value passed by index ("+index+") is greater than or equal to layer size ("+this.neurons.size()+")");
        } else{
            try{
                Neuron neuron=((Neuron)this.neurons.get(index));
                return(neuron);
            } catch(ArrayIndexOutOfBoundsException aiobx){
                throw new LayerException(aiobx.getMessage());
            }
        }
    }

    /** returns the output of the layer
     * @return output of the layer
     */
    public final double[] getOutput(){
        return(this.output);
    }
    
    /** computes the error of the layer (desiredOutput-actualOutput)
     */
    public final void computeError(){
        this.error=0.0D;
        int size=this.neurons.size();
        for(int index=0;index<size;index++){
            this.error+=Math.abs(((Neuron)this.neurons.get(index)).getDesiredOutput()-this.output[index]);
        }
    }
    
    /** returns the error of the layer
     * @return error of the layer     
     */
    public final double getError(){
        return(this.error);
    }
    
    /** returns the number of neurons in the layer
     * @return number of neurons in the layer
     */
    public final int getSize(){
        return(this.neurons.size());
    }
    
    /** returns the output of a specific neuron in the layer
     * @param index the index of the neuron output to return
     * @return the output of the neuron at index    
     * @throws LayerException if index<0 or index>=number of neurons in the layer
     */
    public final double getOutput(int index) throws LayerException{
        double value=0.0D;
        try{
            value=this.output[index];
        } catch(ArrayIndexOutOfBoundsException aiobx){
            throw new LayerException(aiobx.getMessage());
        }
        return(value);
    }

    /** sets the output of a specific neuron in the layer
     * @param index the index of the neuron output to set
     * @param value the new output
     * @throws LayerException if index<0 or index>=number of neurons in the layer
     */    
    public final void setOutput(int index,double value) throws LayerException{
        try{
            this.output[index]=value;
        } catch(ArrayIndexOutOfBoundsException aiobx){
            throw new LayerException(aiobx.getMessage());
        }
    }
    
    /** clears the output array
     */
    public final void initOutput(){
        this.output=new double[this.neurons.size()];
    }
    
    /** returns the training status of the layer
     * @return true if the layer is done training
     */
    public final boolean getDoneTraining(){
        boolean doneTraining=true;
        int size=this.neurons.size();
        int index=0;
        while((doneTraining==true)&&(index<size)){
            doneTraining=((Neuron)this.neurons.get(index)).getDoneTraining();
            index++;
        }
        return(doneTraining);
    }

    /** sets the training status of the layer
     * @param doneTraining set to true if the layer is done training
     */
    public final void setDoneTraining(boolean doneTraining){
        int size=this.neurons.size();
        for(int index=0;index<size;index++){
            Neuron neuron=(Neuron)this.neurons.get(index);
            neuron.setDoneTraining(doneTraining);
            this.neurons.set(index,neuron);
        }
    }

    /** adjusts training rates and error thresholds for every neuron in the layer 
     * values greater than 1.0 increase training rate, values between 0.0 and 1.0 decrease training rate
     * @param rate the new training rate
     */
    public final void adjustTrainingRate(double rate){
        int size=this.neurons.size();
        for(int index=0;index<size;index++){
            Neuron neuron=(Neuron)this.neurons.get(index);
            neuron.setPunishRate(neuron.getPunishRate()*rate);
            neuron.setRewardRate(neuron.getRewardRate()*rate);
            neuron.setErrorThreshold(neuron.getErrorThreshold()*rate);
            neuron.setDoneTraining(false);
            this.neurons.set(index,neuron);
        }
    }

    /** begins thread execution
     */
    public void start(){
        if(myThread==null){
            myThread=new Thread(this);
        }
        myThread.start();
    }

    /** stops thread execution
     */
    public void stop(){
        myThread=null;    
    }        
    
    /* public methods that can be overridden by subclasses */

     /** run the layer
     */
    public void run(){
        int size=this.neurons.size();
        for(int index=0;index<size;index++){
            Neuron neuron=(Neuron)this.neurons.get(index);
            try{
                neuron.run();
            } catch(NeuronException nx){
                nx.printStackTrace();
            }
            this.neurons.set(index,neuron);
        }
        this.computeOutput();
        this.computeError();
        /* notify run is complete */
        this.runComplete();
    }
    
    /** method invoked when run is complete
     */
    public void runComplete(){
        /* override this in implementing class */
    }
    
    /** returns a string representation of the object
     * @return string representation of the object
     */
    public String toString(){
        StringBuffer tostring=new StringBuffer();
        tostring.append(this.getClass().getName());        
        tostring.append("\nerror="+this.error);
        tostring.append("\noutput=");
        int length=this.output.length;
        for(int index=0;index<length;index++){
            tostring.append(this.output[index]+" ");
        }
        tostring.append("\nneurons=");
        length=this.neurons.size();
        for(int index=0;index<length;index++){
            tostring.append("\n"+(this.neurons.get(index)).toString());
        }
        return(tostring.toString());
    }
    
    /* the remaining methods must be overridden by subclasses */

    /** train the layer */
    public abstract void train();
    
    /** compute the output of the layer */
    public abstract void computeOutput();

}