package com.huguesjohnson.jail.neuralNetwork;

/*
neuralNetwork - Package of neural network objects
Neuron.java - Abstract base class for all neurons
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.io.Serializable;
import java.lang.Math;
import java.lang.Object;
import java.lang.String;
import java.lang.StringBuffer;
import com.huguesjohnson.jail.neuralNetwork.NeuronException;

/** class Neuron base class for neuron objects
 * contains a functionality that is common to all neurons
 */
public abstract class Neuron extends Object implements Serializable{
    /** array of input to the neuron */    
    private double[] inputs; 
    
    /** array of weights in network */    
    private double[] weights; 
    
    /** training status of a neuron */    
    private boolean doneTraining;
    
    /** desired output for neuron */
    private double desiredOutput; 
    
    /** punish rate for neuron */
    private double punishRate;
    
    /** reward rate for neuron */
    private double rewardRate;
    
    /** error threshold for neuron */
    private double errorThreshold;
    
    /* constructor(s) */
    
    /** default constructor
     */
    public Neuron(){
        this.inputs=new double[0];
        this.weights=new double[0];
        this.desiredOutput=0.0D;
        this.punishRate=0.01D;
        this.rewardRate=0.01D;
        this.doneTraining=false;
    }
    
    /** constructor
     * @param desiredOutput the desired output of the neuron
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */
    public Neuron(double desiredOutput,double punishRate,double rewardRate,double errorThreshold){
        this.inputs=new double[0];
        this.weights=new double[0];
        this.desiredOutput=desiredOutput;
        this.punishRate=punishRate;
        this.rewardRate=rewardRate;
        this.errorThreshold=errorThreshold;
        this.doneTraining=false;
    }

    /** constructor
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */
    public Neuron(double punishRate,double rewardRate,double errorThreshold){
        this.inputs=new double[0];
        this.weights=new double[0];
        this.desiredOutput=0.0D;
        this.punishRate=punishRate;
        this.rewardRate=rewardRate;
        this.errorThreshold=errorThreshold;
        this.doneTraining=false;
    }    
    
    /** constructor
     * @param input the input to the neuron
     * @param weights the weights on the links to the neuron
     * @param desiredOutput the desired output of the neuron
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */    
    public Neuron(double[] input,double weights[],double desiredOutput,double punishRate,double rewardRate,double errorThreshold){
        this.setInput(input);
        this.setWeights(weights);
        this.desiredOutput=desiredOutput;
        this.punishRate=punishRate;
        this.rewardRate=rewardRate;
        this.errorThreshold=errorThreshold;
        this.doneTraining=false;
    }
    
    /* methods common to all neurons that should not be overridden */
    
    /** getZ computes a summation of all inputs*weights
     * @return summation of all inputs*weights
     */
    public final double getZ(){
        double value=0.0D;
        int length=this.inputs.length;
        if(length==this.weights.length){
            for(int index=0;index<length;index++){
                value+=(this.inputs[index]*this.weights[index]);
            }
        }
        return(value);
    }
    
    /** setInput
     * @param input data for neuron
     */
    public final void setInput(double[] input){
        int length=input.length;
        this.inputs=new double[length];
        for(int index=0;index<length;index++){
            this.inputs[index]=input[index];
        }
    }
    
    /** setWeights
     * @param weights for neuron
     */
    public final void setWeights(double[] weights){
        int length=weights.length;
        this.weights=new double[length];
        for(int index=0;index<length;index++){
            this.weights[index]=weights[index];
        }
    }    
    
    /** getWeights
     * @return weights for neuron
     */
    public final double[] getWeights(){
        return(this.weights);
    }

    /** getInputs same getInput, just around for backwards compatibility
     * @return input to neuron
     */
    protected final double[] getInputs(){
        return(this.getInput());
    }    

    /** getInput 
     * @return input to neuron
     */
    public final double[] getInput(){
        return(this.inputs);
    }     
    
    /** sets random values for weights
     * @param numWeights number of weights to generate
     * @param min minimum weight value
     * @param max maximum weight value
     */
    public final void randomizeWeights(int numWeights,double min,double max){
        this.weights=new double[numWeights];
        for(int index=0;index<numWeights;index++){
            weights[index]=(Math.random()*(max-min))+min;
        }
    }
    
    /** setWeight sets a specific weight in the neuron
     * @param index index of weight to set
     * @param value value to set
     * @throws NeuronException if bad value is passed for index
     */
    public final void setWeight(int index,double value) throws NeuronException{
        if(index<0){
            throw new NeuronException("Value passed by index ("+index+") is less than zero");
        } else if(index>=this.weights.length){
                throw new NeuronException("Value passed by index ("+index+") is greater than or equal to number of weights ("+this.weights.length+")");
        } else{
            this.weights[index]=value;
        }
    }

     /** getWeight retreives a specific weight in the neuron
      * @return value of weight
      * @param index index of weight to get
      * @throws NeuronException if bad index is passed
      */
    public final double getWeight(int index) throws NeuronException{
        if(index<0){
            throw new NeuronException("Value passed by index ("+index+") is less than zero");
        } else if(index>=this.weights.length){
                throw new NeuronException("Value passed by index ("+index+") is greater than or equal to number of weights ("+this.weights.length+")");
        } else{
            return(this.weights[index]);
        }
    }

     /** getInput returns a specific input in the neuron
      * @return value of input
      * @param index index of input to get
      * @throws IndexOutOfBoundsException if bad index is passed
      */
    public final double getInput(int index) throws IndexOutOfBoundsException{
        if(index<0||index>=this.inputs.length){
            throw new IndexOutOfBoundsException();
        } else{
            return(this.inputs[index]);
        }
    }    
    
    /** run computes the value and output of the neuron
     * @return output of neuron
     * @throws NeuronException if number of inputs and weights do not match
     */
    public final double run() throws NeuronException{
        if(this.inputs.length!=this.weights.length){
            throw new NeuronException("Number of inputs and weights do not match\ninputs.length="+this.inputs.length+"\nweights.length="+this.weights.length);
        } else{
            return(this.activation());
        }
    }

    /** run computes the value and output of the neuron
     * @param input array of input for neuron
     * @return output of neuron
     * @throws NeuronException if number of inputs and weights do not match
     */    
    public final double run(double[] input) throws NeuronException{
        this.setInput(input);
        return(this.run());
    }

    /** run computes the value and output of the neuron
     * @param input array of input for neuron
     * @param weights array of weights for neuron
     * @return output of neuron
     * @throws NeuronException if number of inputs and weights do not match
     */    
    public final double run(double[] input,double[] weights) throws NeuronException{
        this.setInput(input);
        this.setWeights(weights);
        return(this.run());
    }    
    
    /** returns whether or not the neuron is done training
     * @return true if the neuron is done training
     */
    public final boolean getDoneTraining(){
        return(this.doneTraining);
    }

     /** sets whether or not the neuron is done training
      * @param doneTraining new training status for neuron
      */
    public final void setDoneTraining(boolean doneTraining){
        this.doneTraining=doneTraining;
    }
    
    /** sets the desired output for the neuron
     * @param desiredOutput the desired output for the neuron
     */
    public final void setDesiredOutput(double desiredOutput){
        this.desiredOutput=desiredOutput;
    }

    /** returns the desired output of the neuron
     * @return desired output of the neuron
     */
     public final double getDesiredOutput(){
        return(this.desiredOutput);
    }    

    /** sets the punish rate for the neuron
     * @param punishRate new punish rate
     */
    public final void setPunishRate(double punishRate){
        this.punishRate=punishRate;
    }

    /** returns the punish rate for the neuron
     * @return punish rate for the neuron
     */
     public final double getPunishRate(){
        return(this.punishRate);
    }        

    /** sets the reward rate for the neuron
     * @param rewardRate new reward rate
     */
    public final void setRewardRate(double rewardRate){
        this.rewardRate=rewardRate;
    }

    /** returns the reward rate for the neuron
     * @return reward rate for the neuron
     */
     public final double getRewardRate(){
        return(this.rewardRate);
    }       
    
    /** sets the error threshold for the neuron
     * @param errorThreshold new error threshold
     */
    public final void setErrorThreshold(double errorThreshold){
        this.errorThreshold=errorThreshold;
    }

    /** returns the error threshold for the neuron
     * @return error threshold for the neuron
     */
     public final double getErrorThreshold(){
        return(this.errorThreshold);
    }    
    
    /** move weight closer to zero 
     * @param index the index of the weight to punish
     */
    public final void punishWeight(int index){
        try{
            double weight=this.weights[index];
            double rate=weight*this.punishRate;
            this.weights[index]-=rate;
        } catch(IndexOutOfBoundsException iobx){
            iobx.printStackTrace();
        }
    }

    /** move weight farther away from zero 
     * @param index the index of the weight to reward
     */
    public final void rewardWeight(int index){
        try{
            double weight=this.weights[index];
            double rate=weight*this.rewardRate;
            this.weights[index]+=rate;
        } catch(IndexOutOfBoundsException iobx){
            iobx.printStackTrace();
        }
    }    
    
    /** returns the number of weights for this neuron
     * @return number of weights for this neuron
     */
    public final int getWeightCount(){
        return(this.weights.length);
    }
    
    /** returns the number of inputs to this neuron
     * @return number of inputs to this neuron
     */
    public final int getInputCount(){
        return(this.inputs.length);
    }    
    
    /** returns the output (activation) of this neuron
     * @return output (activation) of this neuron
     */
    public final double getOutput(){
        return(this.activation());
    }    
    
    /* public methods that can be overridden by subclasses */
    
    /** returns a string representation of the object
     * @return string representation of object
     */
    public String toString(){
        StringBuffer tostring=new StringBuffer();
        tostring.append(this.getClass().getName());
        tostring.append("\ninputs=");
        for(int index=0;index<this.inputs.length;index++){
            tostring.append(String.valueOf(this.inputs[index])+" ");
        }
        tostring.append("\nweights=");
        for(int index=0;index<this.weights.length;index++){
            tostring.append(String.valueOf(this.weights[index])+" ");
        }
        tostring.append("\nz="+String.valueOf(this.getZ()));
        double activation=this.activation();
        tostring.append("\nactivation="+String.valueOf(activation));
        tostring.append("\ndesiredOutput="+String.valueOf(this.desiredOutput));
        tostring.append("\nerrorThreshold="+String.valueOf(this.errorThreshold));
        tostring.append("\nerror="+String.valueOf(Math.abs(desiredOutput-this.activation())));
        tostring.append("\ndoneTraining="+String.valueOf(this.doneTraining));
        return(tostring.toString());
    }    
    
    /* the remaining methods must be overridden by subclasses */
    
    /** activation function for the neuron
     * @return activation value
     */
    public abstract double activation();
    
    /** adjust the weights for this neuron
     */
    public abstract void adjustWeights();
}